import os
import pandas as pd
import numpy as np
import pysam
import pickle
import subprocess
import pybedtools
import shutil
import multiprocessing
import tempfile
import yaml

# Load configuration from config.yaml
with open("config.yaml") as f:
    config = yaml.safe_load(f)

# Extracting paths from configuration
vcf_folder_path = config["vcf_folder_path"]
chip_seq_base_path = config["chip_seq_base_path"]
output_base_path = config["output_base_path"]
reference_genome_path = config["reference_genome_path"]
vcf_error_log_dir = config["vcf_error_log_dir"]
pybedtools.helpers.set_tempdir(config["pybedtools_temp"])

# Load reference genome for sequence extraction (not yet used here)
reference_fasta = pysam.FastaFile(reference_genome_path)

# Delete all contents from temporary directory (e.g. pybedtools tempdir)
def clear_temp_directory(temp_dir):
    for filename in os.listdir(temp_dir):
        file_path = os.path.join(temp_dir, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print(f"Failed to delete {file_path}. Reason: {e}")

# Recursively collect all VCF files, excluding logs directory
def get_vcf_gz_files_except_logs(root_folder):
    all_files = []
    for dirpath, dirnames, filenames in os.walk(root_folder):
        if 'logs' in dirnames:
            dirnames.remove('logs')
        for filename in filenames:
            if filename.endswith('.vcf.gz'):
                all_files.append(os.path.join(dirpath, filename))
    return all_files

# Reindex VCF using tabix for fast querying (optional utility)
def reindex_vcf(vcf_path):
    try:
        subprocess.run(['tabix', '-p', 'vcf', vcf_path], check=True)
        print(f"Re-indexed successfully: {vcf_path}")
    except subprocess.CalledProcessError as e:
        print(f"Failed to re-index {vcf_path}: {e}")

# Convert a VCF file to a DataFrame with coordinates and metadata
def vcf_to_dataframe(vcf_path):
    try:
        vcf_file = pysam.VariantFile(vcf_path)
    except ValueError as e:
        print(f"Failed to open file {vcf_path}: {e}")
        os.makedirs(vcf_error_log_dir, exist_ok=True)
        with open(os.path.join(vcf_error_log_dir, vcf_path.split("/")[-2] + "_error_log.txt"), "a") as error_log:
            error_log.write(f"{vcf_path}\n")
        return None

    data = []
    columns = ["CHROM", "START_POS", "ID", "REF", "ALT", "QUAL", "FILTER"] + list(vcf_file.header.info.keys())
    for record in vcf_file:
        basic_data = [record.chrom, record.pos, record.id, record.ref, ','.join(str(alt) for alt in record.alts), record.qual, record.filter.keys()[0] if record.filter.keys() else 'PASS']
        row_data = [record.info.get(key) for key in vcf_file.header.info.keys()]
        data.append(basic_data + row_data)

    df = pd.DataFrame(data, columns=columns)
    df["START_POS"] = df["START_POS"] - 1  # Convert to 0-based
    df.insert(2, 'END_POS', df["START_POS"] + df['REF'].str.len())
    vcf_file.close()
    return df

# Main function that intersects each TFBS model with all VCFs
def process_tfbs_data(sub_df, temp_dir, files):
    for _, row in sub_df.iterrows():
        tf_name = row[0]
        file_path = f"{chip_seq_base_path}/{tf_name}/300bp_unique_balanced.bed"
        print(file_path)
        tfbs_output_path = f"{output_base_path}/{tf_name}"
        os.makedirs(tfbs_output_path, exist_ok=True)

        df_tfbs = pd.read_csv(file_path, sep="\t")
        TFBS_bed = pybedtools.BedTool.from_dataframe(df_tfbs)
        intersected_vcf_data = {}
        df_statistics = pd.DataFrame(columns=["filename", "Patient_ID", 'work_flow', 'VCF_instance', 'VCF_feature', 'Intersected_instances', 'VCF_column_names'])

        for file_path in files:
            file_name = os.path.basename(file_path)
            parts = file_name.split('.')
            df_vcf = vcf_to_dataframe(file_path)
            if df_vcf is None:
                continue

            vcf_bed = pybedtools.BedTool.from_dataframe(df_vcf)
            intersect_temp_path = os.path.join(temp_dir, "intersect_temp.bed")
            TFBS_bed.intersect(vcf_bed, wa=True, wb=True, output=intersect_temp_path)

            column_list = df_tfbs.columns.to_list() + df_vcf.columns.to_list()
            df_intersection = pybedtools.BedTool(intersect_temp_path).to_dataframe(names=column_list)
            df_intersection = df_intersection[(df_intersection['REF'].str.len() < 10) & (df_intersection['ALT'].str.len() < 10)]

            patient_ID, work_flow = parts[0], parts[2]
            df_statistics.loc[len(df_statistics)] = [file_name, patient_ID, work_flow, df_vcf.shape[0], df_vcf.shape[1], df_intersection.shape[0], list(df_vcf.columns)]
            intersected_vcf_data[f"{patient_ID}_{work_flow}"] = df_intersection

        # Save intersection stats and intersected data per TFBS model
        df_statistics.to_csv(os.path.join(tfbs_output_path, "VCF_statistics.tsv"), sep="\t", index=False)
        with open(os.path.join(tfbs_output_path, "intersected_vcf_data.pkl"), "wb") as file:
            pickle.dump(intersected_vcf_data, file)

# Split a DataFrame into N parts (used for multiprocessing)
def split_dataframe(df, num_splits):
    return np.array_split(df, num_splits)

# Wrapper to call TFBS intersection for a batch in a temp dir
def parallel_process_tfbs_data(sub_df, files):
    with tempfile.TemporaryDirectory(dir=config["pybedtools_temp"]) as temp_dir:
        process_tfbs_data(sub_df, temp_dir, files)

# Entrypoint: given a TFBS model list directory, run the intersection pipeline
def run_tfbs_intersection_pipeline(tfbs_list_dir, start=4, end=46, num_processes=5):
    files = get_vcf_gz_files_except_logs(vcf_folder_path)
    print(f"Total VCFs: {len(files)}")

    for file_index in range(start, end):
        print(f"Processing TFBS list file: output_{file_index}.txt")
        tfbs_list_path = os.path.join(tfbs_list_dir, f"output_{file_index}.txt")
        tfbs_df = pd.read_csv(tfbs_list_path, sep="\t", header=None)

        sub_dfs = split_dataframe(tfbs_df, num_processes)
        pool = multiprocessing.Pool(processes=num_processes)
        pool.starmap(parallel_process_tfbs_data, [(sub_df, files) for sub_df in sub_dfs])
        pool.close()
        pool.join()

        clear_temp_directory(config["pybedtools_temp"])
        print(f"Cleaned up temp after processing file index {file_index}")


"""
Command-line interface for DeepVRegulome.

Usage:
    # Score a single variant
    deepvregulome score --chrom chr1 --pos 3456782 --ref A --alt TA \
        --models CTCFL SP1 MYC --genome hg38.fa

    # Score a VCF file
    deepvregulome score-vcf variants.vcf --models CTCFL SP1 --genome hg38.fa -o results.tsv

    # Score from sequences
    deepvregulome score-seq --ref ATCG... --alt ATCG... --models CTCFL

    # List available models
    deepvregulome list
    deepvregulome list --type TF
    deepvregulome search ZNF
"""

import argparse
import sys


def cmd_score(args):
    from deepvregulome import DVR

    dvr = DVR(genome=args.genome, device=args.device)
    result = dvr.score_variant(
        chrom=args.chrom,
        pos=args.pos,
        ref=args.ref,
        alt=args.alt,
        models=args.models,
        model_type=args.type,
        return_attention=args.attention,
    )

    if args.output:
        result.to_csv(args.output, sep="\t", index=False)
        print(f"Results saved to {args.output}")
    else:
        print(result.to_string(index=False))


def cmd_score_seq(args):
    from deepvregulome import DVR

    dvr = DVR(device=args.device)
    result = dvr.score_sequence(
        ref_seq=args.ref,
        alt_seq=args.alt,
        models=args.models,
        model_type=args.type,
        return_attention=args.attention,
    )

    if args.output:
        result.to_csv(args.output, sep="\t", index=False)
    else:
        print(result.to_string(index=False))


def cmd_score_vcf(args):
    from deepvregulome import DVR

    dvr = DVR(genome=args.genome, device=args.device)
    result = dvr.score_vcf(
        vcf_path=args.vcf,
        models=args.models,
        model_type=args.type,
        return_attention=args.attention,
        max_variants=args.max_variants,
    )

    output = args.output or args.vcf.replace(".vcf", "_dvr_scores.tsv")
    result.to_csv(output, sep="\t", index=False)
    print(f"Scored {len(result)} variant×model combinations → {output}")
    disrupted = result["disrupted"].sum() if "disrupted" in result.columns else 0
    print(f"Disrupted: {disrupted}/{len(result)}")


def cmd_list(args):
    from deepvregulome import ModelRegistry

    reg = ModelRegistry()
    models = reg.list(model_type=args.type)
    print(f"{'Name':<15} {'Type':<8} {'Acc':>6} {'ROC-AUC':>8} {'Peaks':>8}")
    print("-" * 50)
    for m in models:
        print(f"{m.name:<15} {m.model_type:<8} {m.accuracy:>6.1f} {m.roc_auc:>8.2f} {m.peak_count:>8}")
    print(f"\nTotal: {len(models)} models")


def cmd_search(args):
    from deepvregulome import ModelRegistry

    reg = ModelRegistry()
    matches = reg.search(args.query)
    if matches:
        print(f"Models matching '{args.query}': {', '.join(matches)}")
    else:
        print(f"No models found matching '{args.query}'")


def main():
    parser = argparse.ArgumentParser(
        prog="deepvregulome",
        description="DeepVRegulome: Regulatory variant effect prediction",
    )
    sub = parser.add_subparsers(dest="command")

    # --- score ---
    p_score = sub.add_parser("score", help="Score a single variant")
    p_score.add_argument("--chrom", required=True, help="Chromosome (e.g., chr1)")
    p_score.add_argument("--pos", type=int, required=True, help="1-based position")
    p_score.add_argument("--ref", required=True, help="Reference allele")
    p_score.add_argument("--alt", required=True, help="Alternate allele")
    p_score.add_argument("--models", nargs="+", help="Model names (e.g., CTCFL SP1)")
    p_score.add_argument("--type", choices=["TF", "HISTONE"], help="Score all models of this type")
    p_score.add_argument("--genome", required=True, help="Reference genome FASTA")
    p_score.add_argument("--attention", action="store_true", help="Include attention scores")
    p_score.add_argument("--device", default=None, help="cuda or cpu")
    p_score.add_argument("-o", "--output", help="Output TSV file")
    p_score.set_defaults(func=cmd_score)

    # --- score-seq ---
    p_seq = sub.add_parser("score-seq", help="Score from pre-extracted sequences")
    p_seq.add_argument("--ref", required=True, help="Reference sequence (301bp)")
    p_seq.add_argument("--alt", required=True, help="Alternate sequence")
    p_seq.add_argument("--models", nargs="+", help="Model names")
    p_seq.add_argument("--type", choices=["TF", "HISTONE"], help="Score all of type")
    p_seq.add_argument("--attention", action="store_true")
    p_seq.add_argument("--device", default=None)
    p_seq.add_argument("-o", "--output", help="Output TSV file")
    p_seq.set_defaults(func=cmd_score_seq)

    # --- score-vcf ---
    p_vcf = sub.add_parser("score-vcf", help="Score variants from a VCF file")
    p_vcf.add_argument("vcf", help="Input VCF file")
    p_vcf.add_argument("--models", nargs="+", help="Model names")
    p_vcf.add_argument("--type", choices=["TF", "HISTONE"], help="Score all of type")
    p_vcf.add_argument("--genome", required=True, help="Reference genome FASTA")
    p_vcf.add_argument("--attention", action="store_true")
    p_vcf.add_argument("--max-variants", type=int, help="Max variants to score")
    p_vcf.add_argument("--device", default=None)
    p_vcf.add_argument("-o", "--output", help="Output TSV file")
    p_vcf.set_defaults(func=cmd_score_vcf)

    # --- list ---
    p_list = sub.add_parser("list", help="List available models")
    p_list.add_argument("--type", choices=["TF", "HISTONE"], help="Filter by type")
    p_list.set_defaults(func=cmd_list)

    # --- search ---
    p_search = sub.add_parser("search", help="Search model names")
    p_search.add_argument("query", help="Search string (e.g., ZNF, GATA)")
    p_search.set_defaults(func=cmd_search)

    args = parser.parse_args()
    if not args.command:
        parser.print_help()
        sys.exit(1)

    args.func(args)


if __name__ == "__main__":
    main()



"""
DeepVRegulome: DNABERT-based regulatory variant effect prediction.

Usage:
    from deepvregulome import DVR

    dvr = DVR(genome="/path/to/hg38.fa")
    result = dvr.score_variant("chr1", 3456782, "A", "C", models=["CTCFL", "SP1"])
    results = dvr.score_vcf("variants.vcf", model_type="TF", batch_size=100, gpus=[0,1,2,3])
"""

__version__ = "0.1.4"

from deepvregulome.registry import ModelRegistry


def __getattr__(name):
    if name == "DVR":
        from deepvregulome.dvr import DVR
        return DVR
    raise AttributeError(f"module 'deepvregulome' has no attribute {name}")


__all__ = ["DVR", "ModelRegistry", "__version__"]



"""
DVR: Main interface for DeepVRegulome variant effect prediction.

v0.1.4:
    - tqdm progress bars for all long-running operations
    - Parallel sequence extraction (multiprocessing)
    - log2 scoring matching published pipeline
    - Coordinate sanity check (auto-detect 1-based vs 0-based)
    - Multi-GPU model distribution
    - OOM-safe (1 model at a time per GPU)
"""

import math
import warnings
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
from transformers import AutoModelForSequenceClassification, AutoTokenizer

from deepvregulome.registry import ModelInfo, ModelRegistry
from deepvregulome.utils import (
    to_kmer,
    extract_variant_sequences,
    extract_variant_sequences_batch,
    parse_vcf,
    detect_coordinate_system,
)

try:
    from tqdm.auto import tqdm
except ImportError:
    # Fallback: no progress bar, just iterate
    def tqdm(iterable, **kwargs):
        desc = kwargs.get("desc", "")
        total = kwargs.get("total", None)
        if desc and total:
            print(f"  {desc} ({total} items)...")
        return iterable


# ---------------------------------------------------------------------------
# Scoring (matching published DeepVRegulome pipeline)
# ---------------------------------------------------------------------------
def _compute_scores(p_ref: float, p_alt: float) -> dict:
    eps = 1e-7
    p_ref_c = max(eps, min(1 - eps, p_ref))
    p_alt_c = max(eps, min(1 - eps, p_alt))
    lo_ref = math.log2(p_ref_c / (1 - p_ref_c))
    lo_alt = math.log2(p_alt_c / (1 - p_alt_c))
    return {
        "log_odds_ratio": round(lo_ref - lo_alt, 4),
        "score_change": round((p_alt - p_ref) * max(p_ref, p_alt), 6),
        "_log_odds_ref": round(lo_ref, 4),
        "_log_odds_alt": round(lo_alt, 4),
    }


# ---------------------------------------------------------------------------
# GPU worker (multi-GPU mode)
# ---------------------------------------------------------------------------
def _gpu_worker(
    gpu_id, model_names, ref_seqs, alt_seqs,
    batch_size, return_attention, verbose, cache_dir, result_queue,
):
    device = f"cuda:{gpu_id}"
    registry = ModelRegistry()
    all_results = []

    for model_idx, name in enumerate(model_names):
        info = registry.get(name)
        tokenizer = AutoTokenizer.from_pretrained(
            info.hf_repo, subfolder=info.subfolder, cache_dir=cache_dir)
        model = AutoModelForSequenceClassification.from_pretrained(
            info.hf_repo, subfolder=info.subfolder, cache_dir=cache_dir,
            output_attentions=return_attention)
        model = model.to(device)
        model.eval()

        n_batches = math.ceil(len(ref_seqs) / batch_size)
        for batch_idx in range(n_batches):
            batch_start = batch_idx * batch_size
            batch_end = min(batch_start + batch_size, len(ref_seqs))
            batch_refs = ref_seqs[batch_start:batch_end]
            batch_alts = alt_seqs[batch_start:batch_end]

            ref_kmers = [to_kmer(s) for s in batch_refs]
            alt_kmers = [to_kmer(s) for s in batch_alts]

            ref_inputs = tokenizer(ref_kmers, return_tensors="pt", max_length=512,
                                   truncation=True, padding=True)
            alt_inputs = tokenizer(alt_kmers, return_tensors="pt", max_length=512,
                                   truncation=True, padding=True)
            ref_inputs = {k: v.to(device) for k, v in ref_inputs.items()}
            alt_inputs = {k: v.to(device) for k, v in alt_inputs.items()}

            with torch.no_grad():
                ref_out = model(**ref_inputs, output_attentions=return_attention)
                alt_out = model(**alt_inputs, output_attentions=return_attention)

            ref_probs = torch.softmax(ref_out.logits, dim=-1)[:, 1].cpu().numpy()
            alt_probs = torch.softmax(alt_out.logits, dim=-1)[:, 1].cpu().numpy()

            for i in range(len(batch_refs)):
                scores = _compute_scores(float(ref_probs[i]), float(alt_probs[i]))
                row = {
                    "_var_idx": batch_start + i,
                    "model": name, "type": info.model_type,
                    "prob_ref": round(float(ref_probs[i]), 6),
                    "prob_alt": round(float(alt_probs[i]), 6),
                    "log_odds_ratio": scores["log_odds_ratio"],
                    "score_change": scores["score_change"],
                }
                if verbose:
                    row["log_odds_ref"] = scores["_log_odds_ref"]
                    row["log_odds_alt"] = scores["_log_odds_alt"]
                if return_attention and ref_out.attentions and alt_out.attentions:
                    ra = torch.stack(ref_out.attentions)[:, i].cpu().numpy()
                    aa = torch.stack(alt_out.attentions)[:, i].cpu().numpy()
                    diff = aa.mean(axis=1) - ra.mean(axis=1)
                    row["attention_score_change"] = round(float(np.sqrt((diff**2).sum())), 6)
                    row["max_attention_shift"] = round(float(np.abs(diff).max()), 6)
                    lc = np.abs(diff).mean(axis=(1, 2))
                    row["disrupted_layers"] = int((lc > 0.01).sum())
                    row["total_layers"] = len(lc)
                all_results.append(row)

        del model
        torch.cuda.empty_cache()
        print(f"  [GPU {gpu_id}] ✓ {name} ({model_idx + 1}/{len(model_names)})")

    result_queue.put(all_results)


class DVR:
    """
    DeepVRegulome: Score regulatory variant effects using fine-tuned DNABERT models.

    Output columns (default):
        model, type, prob_ref, prob_alt, log_odds_ratio, score_change
    """

    def __init__(
        self,
        genome: Optional[str] = None,
        device: Optional[str] = None,
        cache_dir: Optional[str] = None,
        coordinate_system: Optional[str] = None,
    ):
        self.registry = ModelRegistry()
        self.device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        self.cache_dir = cache_dir
        self._genome_path = genome
        self._genome = None
        self._coordinate_system = coordinate_system
        self._coord_offset = None

    @property
    def genome(self):
        if self._genome is None:
            if self._genome_path is None:
                raise ValueError(
                    "Reference genome not provided. Either:\n"
                    "  1. Pass genome='/path/to/hg38.fa' to DVR()\n"
                    "  2. Use dvr.score_sequence() with pre-extracted sequences"
                )
            try:
                import pysam
            except ImportError:
                raise ImportError(
                    "pysam is required for genome-based scoring. "
                    "Install with: pip install deepvregulome[genome]"
                )
            self._genome = pysam.FastaFile(self._genome_path)
        return self._genome

    def _run_sanity_check(self, variants: list):
        if self._coord_offset is not None:
            return
        if self._coordinate_system == "1-based":
            self._coord_offset = 1
            print("✓ Coordinate system: 1-based (user override). pos → pos-1")
            return
        elif self._coordinate_system == "0-based":
            self._coord_offset = 0
            print("✓ Coordinate system: 0-based (user override). pos used as-is.")
            return
        result = detect_coordinate_system(self.genome, variants)
        self._coord_offset = result["offset"]
        print(result["message"])

    def _adjust_pos(self, pos: int) -> int:
        if self._coord_offset is None:
            return pos - 1
        return pos - self._coord_offset

    # ------------------------------------------------------------------
    # Internal: model loading and prediction
    # ------------------------------------------------------------------
    def _load_model_to_device(self, name: str, device: str):
        info = self.registry.get(name)
        tokenizer = AutoTokenizer.from_pretrained(
            info.hf_repo, subfolder=info.subfolder, cache_dir=self.cache_dir)
        model = AutoModelForSequenceClassification.from_pretrained(
            info.hf_repo, subfolder=info.subfolder, cache_dir=self.cache_dir,
            output_attentions=True)
        model = model.to(device)
        model.eval()
        return model, tokenizer

    def _predict_single(self, model, tokenizer, sequence, device, return_attention=False):
        kmer_seq = to_kmer(sequence, k=6)
        inputs = tokenizer(kmer_seq, return_tensors="pt", max_length=512,
                           truncation=True, padding=True)
        inputs = {k: v.to(device) for k, v in inputs.items()}
        with torch.no_grad():
            outputs = model(**inputs, output_attentions=return_attention)
        probs = torch.softmax(outputs.logits, dim=-1)
        result = {"prob": probs[0][1].item()}
        if return_attention and outputs.attentions:
            result["attention"] = torch.stack(outputs.attentions).squeeze(1).cpu().numpy()
        return result

    def _predict_batch(self, model, tokenizer, sequences, device, return_attention=False):
        kmer_seqs = [to_kmer(s) for s in sequences]
        inputs = tokenizer(kmer_seqs, return_tensors="pt", max_length=512,
                           truncation=True, padding=True)
        inputs = {k: v.to(device) for k, v in inputs.items()}
        with torch.no_grad():
            outputs = model(**inputs, output_attentions=return_attention)
        probs = torch.softmax(outputs.logits, dim=-1)[:, 1].cpu().numpy()
        results = []
        for i in range(len(sequences)):
            r = {"prob": float(probs[i])}
            if return_attention and outputs.attentions:
                r["attention"] = torch.stack(outputs.attentions)[:, i].cpu().numpy()
            results.append(r)
        return results

    def _compute_attention_change(self, attn_ref, attn_alt):
        diff = attn_alt.mean(axis=1) - attn_ref.mean(axis=1)
        return {
            "attention_score_change": round(float(np.sqrt((diff**2).sum())), 6),
            "max_attention_shift": round(float(np.abs(diff).max()), 6),
            "disrupted_layers": int((np.abs(diff).mean(axis=(1, 2)) > 0.01).sum()),
            "total_layers": diff.shape[0],
        }

    # ------------------------------------------------------------------
    # Internal: single-GPU scoring (memory-safe, with tqdm)
    # ------------------------------------------------------------------
    def _score_sequences_single_gpu(
        self, ref_seqs, alt_seqs, model_names,
        batch_size=1, return_attention=False, verbose=False, device=None,
    ):
        device = device or self.device
        results = []

        model_pbar = tqdm(model_names, desc="Models", unit="model")
        for model_idx, name in enumerate(model_pbar):
            model_pbar.set_postfix(current=name)
            info = self.registry.get(name)
            model, tokenizer = self._load_model_to_device(name, device)

            n_batches = math.ceil(len(ref_seqs) / batch_size)
            batch_pbar = tqdm(
                range(n_batches),
                desc=f"  {name}",
                unit="batch",
                leave=False,
            )

            for batch_idx in batch_pbar:
                batch_start = batch_idx * batch_size
                batch_end = min(batch_start + batch_size, len(ref_seqs))
                br = ref_seqs[batch_start:batch_end]
                ba = alt_seqs[batch_start:batch_end]

                if batch_size == 1:
                    ro = self._predict_single(model, tokenizer, br[0], device, return_attention)
                    ao = self._predict_single(model, tokenizer, ba[0], device, return_attention)
                    rp, ap = [ro["prob"]], [ao["prob"]]
                    ra_list = [ro.get("attention")] if return_attention else [None]
                    aa_list = [ao.get("attention")] if return_attention else [None]
                else:
                    ros = self._predict_batch(model, tokenizer, br, device, return_attention)
                    aos = self._predict_batch(model, tokenizer, ba, device, return_attention)
                    rp = [r["prob"] for r in ros]
                    ap = [r["prob"] for r in aos]
                    ra_list = [r.get("attention") for r in ros] if return_attention else [None]*len(ros)
                    aa_list = [r.get("attention") for r in aos] if return_attention else [None]*len(aos)

                for i in range(len(br)):
                    scores = _compute_scores(rp[i], ap[i])
                    row = {
                        "_var_idx": batch_start + i,
                        "model": name, "type": info.model_type,
                        "prob_ref": round(rp[i], 6),
                        "prob_alt": round(ap[i], 6),
                        "log_odds_ratio": scores["log_odds_ratio"],
                        "score_change": scores["score_change"],
                    }
                    if verbose:
                        row["log_odds_ref"] = scores["_log_odds_ref"]
                        row["log_odds_alt"] = scores["_log_odds_alt"]
                    if return_attention and ra_list[i] is not None and aa_list[i] is not None:
                        row.update(self._compute_attention_change(ra_list[i], aa_list[i]))
                    results.append(row)

            del model, tokenizer
            torch.cuda.empty_cache()

        return results

    # ------------------------------------------------------------------
    # Internal: multi-GPU scoring
    # ------------------------------------------------------------------
    def _score_sequences_multi_gpu(
        self, ref_seqs, alt_seqs, model_names,
        gpus, batch_size=32, return_attention=False, verbose=False,
    ):
        import torch.multiprocessing as mp

        n_gpus = len(gpus)
        splits = []
        for i in range(n_gpus):
            s = i * len(model_names) // n_gpus
            e = (i + 1) * len(model_names) // n_gpus
            splits.append(model_names[s:e])

        print(f"Distributing {len(model_names)} models across {n_gpus} GPUs: "
              f"{[len(s) for s in splits]} models each")

        mp.set_start_method("spawn", force=True)
        result_queue = mp.Queue()
        processes = []

        for i, gpu_id in enumerate(gpus):
            if not splits[i]:
                continue
            p = mp.Process(target=_gpu_worker, args=(
                gpu_id, splits[i], ref_seqs, alt_seqs,
                batch_size, return_attention, verbose, self.cache_dir,
                result_queue))
            p.start()
            processes.append(p)

        all_results = []
        for _ in processes:
            all_results.extend(result_queue.get())
        for p in processes:
            p.join()

        return all_results

    # ==================================================================
    # PUBLIC API
    # ==================================================================
    def score_sequence(
        self,
        ref_seq: Union[str, List[str]],
        alt_seq: Union[str, List[str]],
        models=None, model_type=None,
        batch_size: int = 1, gpus=None,
        return_attention: bool = False,
        verbose: bool = False,
    ) -> pd.DataFrame:
        """Score variant(s) from pre-extracted REF and ALT sequences."""
        if isinstance(ref_seq, str):
            ref_seqs, alt_seqs = [ref_seq], [alt_seq]
        else:
            ref_seqs, alt_seqs = list(ref_seq), list(alt_seq)

        model_names = self._resolve_models(models, model_type)

        if gpus and len(gpus) > 1:
            raw = self._score_sequences_multi_gpu(
                ref_seqs, alt_seqs, model_names,
                gpus=gpus, batch_size=batch_size,
                return_attention=return_attention, verbose=verbose)
        else:
            device = f"cuda:{gpus[0]}" if gpus else self.device
            raw = self._score_sequences_single_gpu(
                ref_seqs, alt_seqs, model_names,
                batch_size=batch_size, return_attention=return_attention,
                verbose=verbose, device=device)

        df = pd.DataFrame(raw)
        if "_var_idx" in df.columns:
            df = df.drop(columns=["_var_idx"])
        if len(df) > 0:
            df = df.sort_values("log_odds_ratio", ascending=False, key=abs)
        return df.reset_index(drop=True)

    def score_variant(
        self, chrom: str, pos: int, ref: str, alt: str,
        models=None, model_type=None, flank: int = 150,
        batch_size: int = 1, gpus=None,
        return_attention: bool = False, verbose: bool = False,
    ) -> pd.DataFrame:
        """Score a single variant by genomic coordinates."""
        self._run_sanity_check([{"chrom": chrom, "pos": pos, "ref": ref, "alt": alt}])
        pos_0 = self._adjust_pos(pos)
        ref_seq, alt_seq = extract_variant_sequences(
            self.genome, chrom, pos_0, ref, alt, flank=flank)
        df = self.score_sequence(
            ref_seq, alt_seq, models=models, model_type=model_type,
            batch_size=batch_size, gpus=gpus,
            return_attention=return_attention, verbose=verbose)
        df.insert(0, "chrom", chrom)
        df.insert(1, "pos", pos)
        df.insert(2, "ref", ref)
        df.insert(3, "alt", alt)
        return df

    def score_variants(
        self, variants: pd.DataFrame,
        models=None, model_type=None, flank: int = 150,
        batch_size: int = 32, gpus=None,
        return_attention: bool = False, verbose: bool = False,
    ) -> pd.DataFrame:
        """Score multiple variants from a DataFrame (columns: chrom, pos, ref, alt)."""
        required = {"chrom", "pos", "ref", "alt"}
        if not required.issubset(variants.columns):
            raise ValueError(f"Missing columns: {required - set(variants.columns)}")

        variant_list = variants[["chrom", "pos", "ref", "alt"]].to_dict("records")

        # Sanity check
        self._run_sanity_check(variant_list)

        # Adjust positions to 0-based
        adjusted = []
        for v in variant_list:
            adjusted.append({
                "chrom": v["chrom"],
                "pos": self._adjust_pos(v["pos"]),
                "ref": v["ref"],
                "alt": v["alt"],
            })

        # Parallel sequence extraction with tqdm
        print(f"Extracting sequences for {len(variants)} variants...")
        seq_pairs = extract_variant_sequences_batch(
            self._genome_path, adjusted, flank
        )

        valid_indices, ref_seqs, alt_seqs = [], [], []
        for i, (r, a) in enumerate(seq_pairs):
            if r is not None and a is not None:
                valid_indices.append(i)
                ref_seqs.append(r)
                alt_seqs.append(a)

        if len(valid_indices) < len(variants):
            n_failed = len(variants) - len(valid_indices)
            warnings.warn(f"Failed to extract sequences for {n_failed} variants")
        if not ref_seqs:
            return pd.DataFrame()

        model_names = self._resolve_models(models, model_type)
        print(f"Scoring {len(ref_seqs)} variants × {len(model_names)} models...")

        if gpus and len(gpus) > 1:
            raw = self._score_sequences_multi_gpu(
                ref_seqs, alt_seqs, model_names,
                gpus=gpus, batch_size=batch_size,
                return_attention=return_attention, verbose=verbose)
        else:
            device = f"cuda:{gpus[0]}" if gpus else self.device
            raw = self._score_sequences_single_gpu(
                ref_seqs, alt_seqs, model_names,
                batch_size=batch_size, return_attention=return_attention,
                verbose=verbose, device=device)

        df = pd.DataFrame(raw)
        if len(df) > 0 and "_var_idx" in df.columns:
            var_info = variants.iloc[valid_indices].reset_index(drop=True)
            df["chrom"] = df["_var_idx"].map(lambda i: var_info.loc[i, "chrom"])
            df["pos"] = df["_var_idx"].map(lambda i: var_info.loc[i, "pos"])
            df["ref"] = df["_var_idx"].map(lambda i: var_info.loc[i, "ref"])
            df["alt"] = df["_var_idx"].map(lambda i: var_info.loc[i, "alt"])
            df = df.drop(columns=["_var_idx"])
            cols = ["chrom", "pos", "ref", "alt"] + [c for c in df.columns if c not in ["chrom", "pos", "ref", "alt"]]
            df = df[cols]
            df = df.sort_values("log_odds_ratio", ascending=False, key=abs)
        return df.reset_index(drop=True)

    def score_vcf(
        self, vcf_path: str,
        models=None, model_type=None, flank: int = 150,
        batch_size: int = 32, gpus=None,
        return_attention: bool = False, verbose: bool = False,
        max_variants: Optional[int] = None,
    ) -> pd.DataFrame:
        """Score all variants in a VCF file."""
        print(f"Parsing VCF: {vcf_path}")
        variant_list = parse_vcf(vcf_path, max_variants=max_variants)
        print(f"  Found {len(variant_list)} variants")
        if not variant_list:
            return pd.DataFrame()
        return self.score_variants(
            pd.DataFrame(variant_list), models=models, model_type=model_type,
            flank=flank, batch_size=batch_size, gpus=gpus,
            return_attention=return_attention, verbose=verbose)

    # ==================================================================
    # Helpers
    # ==================================================================
    def _resolve_models(self, models, model_type):
        if models and model_type:
            raise ValueError("Specify either 'models' or 'model_type', not both")
        if models:
            for m in models:
                self.registry.get(m)
            return models
        if model_type:
            return [m.name for m in self.registry.list(model_type=model_type)]
        raise ValueError("Must specify either 'models' or 'model_type' ('TF'/'HISTONE')")

    def list_models(self, model_type=None):
        return self.registry.list(model_type=model_type)

    def search_models(self, query: str):
        return self.registry.search(query)

    def __repr__(self):
        g = f", genome='{self._genome_path}'" if self._genome_path else ""
        cs = f", coords={self._coordinate_system}" if self._coordinate_system else ""
        return f"DVR(device='{self.device}'{g}{cs}, {self.registry})"

    def __del__(self):
        if self._genome is not None:
            try:
                self._genome.close()
            except Exception:
                pass



import os
import pandas as pd
import pysam
import gzip
import pickle, subprocess
import pybedtools
print(pybedtools.__file__)
pybedtools.helpers.set_tempdir('/home/pdutta/pybedtools_temp') ## Set a temporary folder


def get_vcf_gz_files_except_logs(root_folder):
    all_files = []

    # Walk through the directory tree
    for dirpath, dirnames, filenames in os.walk(root_folder):
        # If "logs" is in dirnames, remove it to avoid traversing it
        if 'logs' in dirnames:
            dirnames.remove('logs')

        # Add only the filenames with the extension .vcf.gz in the current directory to the all_files list
        for filename in filenames:
            if filename.endswith('.vcf.gz'):
                all_files.append(os.path.join(dirpath, filename))
    return all_files



def vcf_to_dataframe(file_info):
    """
    Convert a .vcf.gz file into a pandas DataFrame.

    Parameters:
    - file_info (tuple): contains (vcf_path, patient_id)

    Returns:
    - pd.DataFrame: VCF data as a DataFrame with patient_id column
    """
    vcf_path, patient_id = file_info
    print(patient_id)

    try:
        # Open the VCF file
        vcf_file = pysam.VariantFile(vcf_path)
    except ValueError as e:
        print(f"Failed to open file {vcf_path}: {e}")
        with open("/data/private/pdutta_new/GDC_Cancer_Wise/New_data/Brain/" + vcf_path.split("/")[-2] + "_error_log.txt", "a") as error_log:
            error_log.write(f"{vcf_path}\n")
        return None

    # Extracting the data and the columns
    data = []
    columns = ["CHROM", "START_POS", "ID", "REF", "ALT", "QUAL", "FILTER"] + list(vcf_file.header.info.keys())
    for record in vcf_file:
        basic_data = [record.chrom, record.pos, record.id, record.ref,
                      ','.join(str(alt) for alt in record.alts), record.qual,
                      record.filter.keys()[0] if record.filter.keys() else 'PASS']
        row_data = [record.info.get(key) for key in vcf_file.header.info.keys()]
        data.append(basic_data + row_data)

    df = pd.DataFrame(data, columns=columns)
    df["START_POS"] = df["START_POS"] - 1
    end = df["START_POS"] + df['REF'].str.len()
    df.insert(2, 'END_POS', end)

    # Add the "patient_id" column (optional )
    df['patient_id'] = patient_id

    # Close the VCF file
    vcf_file.close()

    return df


def process_files_in_parallel(files):
    # Set up multiprocessing
    with Pool(cpu_count() - 1) as pool:
        results = pool.map(vcf_to_dataframe, files)
    
    # Filter out any None results and concatenate the DataFrames
    return pd.concat([df for df in results if df is not None], ignore_index=True)


"""
Sequence processing utilities for DeepVRegulome.

Handles DNA ↔ k-mer conversion, reverse complement,
variant sequence extraction (with multiprocessing),
VCF parsing, and coordinate sanity checking.
"""

import os
from typing import List, Tuple, Optional
from multiprocessing import Pool


COMPLEMENT = str.maketrans("ACGTNacgtn", "TGCANtgcan")


def to_kmer(sequence: str, k: int = 6) -> str:
    """Convert a DNA sequence to space-separated k-mer representation for DNABERT."""
    seq = sequence.upper()
    return " ".join(seq[i:i + k] for i in range(len(seq) - k + 1))


def reverse_complement(sequence: str) -> str:
    """Return the reverse complement of a DNA sequence."""
    return sequence.translate(COMPLEMENT)[::-1]


def extract_variant_sequences(
    genome,
    chrom: str,
    pos_0based: int,
    ref: str,
    alt: str,
    flank: int = 150,
) -> Tuple[str, str]:
    """
    Extract REF and ALT sequences centered on a variant.

    Args:
        genome: pysam.FastaFile
        chrom: Chromosome (e.g., "chr1")
        pos_0based: 0-based variant position
        ref: Reference allele
        alt: Alternate allele
        flank: Flanking bases on each side

    Returns:
        (ref_seq, alt_seq) tuple
    """
    left_flank = genome.fetch(chrom, pos_0based - flank, pos_0based)
    right_flank = genome.fetch(chrom, pos_0based + len(ref), pos_0based + len(ref) + flank)

    ref_seq = left_flank + ref + right_flank
    alt_seq = left_flank + alt + right_flank

    return ref_seq.upper(), alt_seq.upper()


# ---------------------------------------------------------------------------
# Parallel sequence extraction
# ---------------------------------------------------------------------------

# Module-level global for multiprocessing workers
_worker_genome = None
_worker_flank = None


def _init_extraction_worker(genome_path: str, flank: int):
    """Initialize pysam.FastaFile in each worker process."""
    global _worker_genome, _worker_flank
    import pysam
    _worker_genome = pysam.FastaFile(genome_path)
    _worker_flank = flank


def _extract_one_variant(variant: dict) -> Tuple[Optional[str], Optional[str]]:
    """Extract sequences for a single variant (called by worker)."""
    global _worker_genome, _worker_flank
    try:
        ref_seq, alt_seq = extract_variant_sequences(
            _worker_genome,
            variant["chrom"],
            variant["pos"],
            variant["ref"],
            variant["alt"],
            _worker_flank,
        )
        return (ref_seq, alt_seq)
    except Exception:
        return (None, None)


def extract_variant_sequences_batch(
    genome_path: str,
    variants: list,
    flank: int = 150,
    n_workers: int = 0,
) -> List[Tuple[Optional[str], Optional[str]]]:
    """
    Extract REF and ALT sequences for a batch of variants.
    Uses multiprocessing when n_workers > 1 for speed.

    Args:
        genome_path: Path to reference genome FASTA (string, not pysam object)
        variants: List of dicts with chrom, pos, ref, alt (pos already 0-based)
        flank: Flanking bases on each side
        n_workers: Number of parallel workers. 0 = auto (cpu_count/2, min 1).

    Returns:
        List of (ref_seq, alt_seq) tuples. Failed extractions return (None, None).
    """
    if n_workers == 0:
        n_workers = max(1, os.cpu_count() // 2)

    if n_workers == 1 or len(variants) < 100:
        # Single-threaded for small inputs
        import pysam
        genome = pysam.FastaFile(genome_path)
        results = []
        for v in variants:
            try:
                ref_seq, alt_seq = extract_variant_sequences(
                    genome, v["chrom"], v["pos"], v["ref"], v["alt"], flank
                )
                results.append((ref_seq, alt_seq))
            except Exception:
                results.append((None, None))
        genome.close()
        return results

    # Multi-threaded
    with Pool(
        processes=n_workers,
        initializer=_init_extraction_worker,
        initargs=(genome_path, flank),
    ) as pool:
        results = pool.map(_extract_one_variant, variants, chunksize=500)

    return results


def parse_vcf(vcf_path: str, max_variants: Optional[int] = None) -> list:
    """
    Parse a VCF file into a list of variant dicts.
    Handles both with-header and headerless VCFs.

    Returns:
        List of dicts with keys: chrom, pos, ref, alt
    """
    import gzip

    opener = gzip.open if vcf_path.endswith(".gz") else open
    variants = []

    with opener(vcf_path, "rt") as f:
        for line in f:
            if line.startswith("#"):
                continue

            fields = line.strip().split("\t")
            if len(fields) < 5:
                continue

            chrom = fields[0]
            try:
                pos = int(fields[1])
            except ValueError:
                continue
            ref = fields[3]
            alt_str = fields[4]

            for alt in alt_str.split(","):
                alt = alt.strip()
                if alt == "." or alt == "*":
                    continue
                variants.append({
                    "chrom": chrom,
                    "pos": pos,
                    "ref": ref,
                    "alt": alt,
                })

            if max_variants and len(variants) >= max_variants:
                break

    return variants[:max_variants] if max_variants else variants


# ──────────────────────────────────────────────────────────────
# COORDINATE SANITY CHECK
# ──────────────────────────────────────────────────────────────

def detect_coordinate_system(genome, variants: list, n_check: int = 10) -> dict:
    """
    Auto-detect whether variant positions are 1-based (VCF) or 0-based (BED)
    by checking REF alleles against the reference genome.
    """
    to_check = variants[:min(n_check, len(variants))]

    votes_1based = 0
    votes_0based = 0
    ambiguous = 0
    no_match = 0

    for v in to_check:
        chrom = v["chrom"]
        pos = v["pos"]
        ref = v["ref"].upper()

        try:
            if len(ref) > 1:
                seq_at_minus1 = genome.fetch(chrom, pos - 1, pos - 1 + len(ref)).upper()
                seq_at_pos = genome.fetch(chrom, pos, pos + len(ref)).upper()
            else:
                seq_at_minus1 = genome.fetch(chrom, pos - 1, pos).upper()
                seq_at_pos = genome.fetch(chrom, pos, pos + 1).upper()

            match_minus1 = (seq_at_minus1 == ref)
            match_pos = (seq_at_pos == ref)

            if match_minus1 and not match_pos:
                votes_1based += 1
            elif match_pos and not match_minus1:
                votes_0based += 1
            elif match_minus1 and match_pos:
                ambiguous += 1
            else:
                no_match += 1

        except Exception:
            no_match += 1

    total_checked = len(to_check)

    if votes_1based >= 3 and votes_0based == 0:
        system, offset, confidence = "1-based", 1, "high"
    elif votes_0based >= 3 and votes_1based == 0:
        system, offset, confidence = "0-based", 0, "high"
    elif votes_1based > votes_0based:
        system, offset, confidence = "1-based", 1, "medium"
    elif votes_0based > votes_1based:
        system, offset, confidence = "0-based", 0, "medium"
    else:
        system, offset, confidence = "1-based", 1, "low"

    if confidence == "high":
        message = (
            f"✓ Sanity check PASSED — {system} coordinates detected\n"
            f"  Checked {total_checked} variants: "
            f"{votes_1based} confirm 1-based, {votes_0based} confirm 0-based, "
            f"{ambiguous} ambiguous, {no_match} no-match\n"
            f"  REF alleles match the reference genome. Coordinates look correct."
        )
    elif confidence == "medium":
        message = (
            f"⚠ Sanity check: likely {system} but not fully certain\n"
            f"  Checked {total_checked} variants: "
            f"{votes_1based} suggest 1-based, {votes_0based} suggest 0-based, "
            f"{ambiguous} ambiguous, {no_match} no-match\n"
            f"  Proceeding with {system}. Override with coordinate_system= if needed."
        )
    else:
        message = (
            f"⚠ Sanity check: could not confidently determine coordinate system\n"
            f"  Checked {total_checked} variants: "
            f"{votes_1based} suggest 1-based, {votes_0based} suggest 0-based, "
            f"{ambiguous} ambiguous, {no_match} no-match\n"
            f"  Defaulting to 1-based (VCF standard). Override with coordinate_system='0-based' if needed."
        )

    return {
        "system": system,
        "offset": offset,
        "confidence": confidence,
        "votes_1based": votes_1based,
        "votes_0based": votes_0based,
        "ambiguous": ambiguous,
        "no_match": no_match,
        "message": message,
    }



import os, glob
import pandas as pd
import pysam
import numpy as np
import pickle
from multiprocessing import Pool, cpu_count

# === Configurable paths ===
cancer_type = "Brain"
intersected_base_path = f"/data/private/pdutta_new/GDC_Cancer_Wise/New_data/{cancer_type}/Generated_files/Intersected_Data/Somatic/300bp_TFBS"
reference_genome_path = "/data/projects/Resources/Gencode_genome_annotation/GRCh38.primary_assembly.genome.fa"
output_path = f"/data/private/pdutta_new/GDC_Cancer_Wise/New_data/{cancer_type}/Generated_files/DNABERT_Data/Somatic/TFBS_300bp_new"

os.makedirs(output_path, exist_ok=True)

# === Load top models list ===
top_models_df = pd.read_csv("/home/campus.stonybrook.edu/pdutta/Github/Postdoc/DNABERT_data_processing/TFBS/300bp_TFBS_accuracy_Stat.tsv", sep="\t")
top_models_df = top_models_df[top_models_df['eval_acc'] >= 0.85].iloc[100:200]

# === Convert sequence to space-separated k-mers ===
def seq2kmer(seq, k=6):
    return " ".join([seq[i:i+k] for i in range(len(seq)-k+1)])

# === Apply all mutations to a given reference sequence ===
def apply_mutations_absolute(ref_seq, chip_start, starts, ends, refs, alts):
    mutations = sorted(zip(starts, ends, refs, alts), key=lambda x: x[0])
    seq_list = list(ref_seq)
    offset = chip_start
    for start, end, ref, alt in mutations:
        rel_pos = start - offset
        if ''.join(seq_list[rel_pos:rel_pos+len(ref)]) == ref:
            seq_list[rel_pos:rel_pos+len(ref)] = list(alt)
    return ''.join(seq_list)

# === Load genome reference once per worker ===
def initialize_worker():
    global reference_fasta
    reference_fasta = pysam.FastaFile(reference_genome_path)
    print("##### Reference genome loaded in worker #####")

# === Process a dataframe of variants for a single TFBS-patient combo ===
def get_sequences(df):
    global reference_fasta
    try:
        data = []
        for idx, row in df.iterrows():
            chrom, ref_start, ref_end = row[0], row[1], row[2]
            variant_start, variant_end = row['START_POS'], row['END_POS']
            ref_nucleotide, alt = row['REF'], row['ALT']
            ref_seq = reference_fasta.fetch(chrom, ref_start, ref_end)
            variant_pos_start = variant_start - ref_start
            variant_pos_end = variant_end - ref_start

            if len(ref_nucleotide) < len(alt):  # Insertion
                delete_size = len(alt) - len(ref_nucleotide)
                alt_seq = ref_seq[:variant_pos_start] + alt + ref_seq[variant_pos_end:len(ref_seq)-delete_size]
            elif len(ref_nucleotide) > len(alt):  # Deletion
                insert_size = len(ref_nucleotide) - len(alt)
                extra_bases = reference_fasta.fetch(chrom, ref_end, ref_end+insert_size)
                alt_seq = ref_seq[:variant_pos_start] + alt + ref_seq[variant_pos_end:] + extra_bases
            else:  # SNV
                alt_seq = ref_seq[:variant_pos_start] + alt + ref_seq[variant_pos_end:]

            data.append({
                'chr': chrom, 'Chip_Seq_start': ref_start, 'Chip_Seq_end': ref_end,
                'varinat_start': variant_start, 'variant_end': variant_end,
                'ref_neucleotide': ref_nucleotide, 'alternative_neucleotide': alt,
                'reference_seq': ref_seq, 'alt_seq': alt_seq
            })

        new_df = pd.DataFrame(data).drop_duplicates().reset_index(drop=True)

        merged_list = list(zip(new_df['reference_seq'], new_df['alt_seq']))
        merged_list = [item.upper() for tup in merged_list for item in tup]
        df_kmer = pd.DataFrame(list(map(seq2kmer, merged_list)), columns=['Sequence'])
        df_kmer['Label'] = np.random.choice([0, 1], size=len(df_kmer))

        grouped_df = new_df.groupby(['chr', 'Chip_Seq_start', 'Chip_Seq_end']).agg({
            'varinat_start': list, 'variant_end': list,
            'ref_neucleotide': list, 'alternative_neucleotide': list,
            'reference_seq': 'first', 'alt_seq': lambda x: list(set(x))
        }).reset_index()

        grouped_df['mutated_sequence'] = grouped_df.apply(lambda row: apply_mutations_absolute(
            row['reference_seq'], row['Chip_Seq_start'], row['varinat_start'],
            row['variant_end'], row['ref_neucleotide'], row['alternative_neucleotide']), axis=1)

        merged_regionwise = list(zip(grouped_df['reference_seq'], grouped_df['mutated_sequence']))
        merged_regionwise = [item.upper() for tup in merged_regionwise for item in tup]
        df_kmer_region = pd.DataFrame(list(map(seq2kmer, merged_regionwise)), columns=['Sequence'])
        df_kmer_region['Label'] = np.random.choice([0, 1], size=len(df_kmer_region))

        return new_df, df_kmer, grouped_df, df_kmer_region

    except Exception as e:
        print(f"Error in get_sequences: {e}")
        raise

# === Main processing logic for each TFBS tag ===
def process_row(row):
    global reference_fasta
    try:
        print(f"Processing {row['tags']}")
        intersected_data = f"{intersected_base_path}/{row['tags']}/intersected_vcf_data.pkl"

        with open(intersected_data, "rb") as file:
            loaded_dictionary = pickle.load(file)

        dnabert_raw_data = {}
        dnabert_raw_data_regionwise = {}

        for key, value in loaded_dictionary.items():
            try:
                print(f"  → {key}")
                new_df, df_kmer, grouped_df, df_kmer_region = get_sequences(value)
                dnabert_raw_data[key] = new_df
                dnabert_raw_data_regionwise[key] = grouped_df

                patient_folder = f"{output_path}/{row['tags']}/Patient_wise"
                os.makedirs(f"{patient_folder}/variant_wise/{key}", exist_ok=True)
                os.makedirs(f"{patient_folder}/region_wise/{key}", exist_ok=True)
                df_kmer.to_csv(f"{patient_folder}/variant_wise/{key}/dev.tsv", sep="\t", index=False)
                df_kmer_region.to_csv(f"{patient_folder}/region_wise/{key}/dev.tsv", sep="\t", index=False)

            except Exception as e:
                print(f"Error: {e} in {row['tags']} → {key}")
                return row['tags']

        with open(f"{output_path}/{row['tags']}/variantwise_raw_vcf_data.pkl", "wb") as file:
            pickle.dump(dnabert_raw_data, file)
        with open(f"{output_path}/{row['tags']}/regionwise_raw_vcf_data.pkl", "wb") as file:
            pickle.dump(dnabert_raw_data_regionwise, file)

    except Exception as e:
        print(f"Global error in {row['tags']}: {e}")
        return row['tags']

    return None

# === Run multiprocessing ===
if __name__ == '__main__':
    print(f"Using {cpu_count()} CPUs for processing")
    with Pool(cpu_count() - 10, initializer=initialize_worker) as pool:
        missing_files = pool.map(process_row, [row for _, row in top_models_df.iterrows()])

    missing_files = [tag for tag in missing_files if tag is not None]
    with open(f"{output_path}/missing_files.txt", "w") as f:
        for tag in missing_files:
            f.write(f"{tag}\n")

    print(f"Missing files recorded at {output_path}/missing_files.txt")
